use axum::{
    extract::{Path, State},
    http::HeaderMap,
    http::StatusCode,
    routing::get,
    routing::post,
    routing::put,
    Json, Router,
};
use serde::{Deserialize, Serialize};
use serde_json::{json, Value};
use sqlx::{postgres::PgPoolOptions, PgExecutor, PgPool};

use std::env;
use std::time::{Duration, SystemTime, UNIX_EPOCH};

#[derive(Deserialize)]
pub struct User {
    username: String,
    password: String,
}

#[derive(Deserialize)]
pub struct UpdateProgress {
    document: String,
    progress: String,
    percentage: f32,
    device: String,
    device_id: String,
}

#[derive(Deserialize, Serialize)]
pub struct GetProgress {
    document: String,
    progress: String,
    percentage: f32,
    device: String,
    device_id: String,
    timestamp: i64,
}

const VERSION: &str = env!("CARGO_PKG_VERSION");

#[tokio::main]
async fn main() {
    let pg_url: String = env::var("PG_URL")
        .ok()
        .and_then(|v| v.parse().ok())
        .unwrap();
    let listen_port: String = env::var("PORT").ok().and_then(|v| v.parse().ok()).unwrap();

    let db_pool = PgPoolOptions::new()
        .max_connections(64)
        .acquire_timeout(Duration::from_secs(5))
        .connect(&pg_url)
        .await
        .expect("Can't connect to database.");

    // build our application with a single route
    let app = Router::new()
        .route("/", get(root))
        .route("/users/create", post(create_user))
        .route("/users/auth", get(auth_user))
        .route("/syncs/progress", put(update_progress))
        .route("/syncs/progress/{document}", get(get_progress))
        .route("/healthcheck", get(healthcheck))
        .route("/version", get(version))
        .with_state(db_pool);

    let mut bind_with = "127.0.0.1:".to_owned();
    bind_with += &listen_port;
    let listener = tokio::net::TcpListener::bind(bind_with)
        .await
        .unwrap();
    axum::serve(listener, app).await.unwrap();
}

async fn root() -> &'static str {
    "KOreader sync server"
}

async fn version() -> &'static str {
    VERSION
}

async fn create_user(
    State(db_pool): State<PgPool>,
    Json(payload): Json<User>,
) -> (StatusCode, Json<Value>) {
    //let client = redis::Client::open("redis://127.0.0.1/").unwrap();
    //let mut con = client.get_connection().unwrap();

    let username = payload.username;
    let password = payload.password;

    let row: (i64,) = sqlx::query_as("SELECT COUNT(id) FROM users WHERE username = $1")
        .bind(&username)
        .fetch_one(&db_pool)
        .await
        .unwrap();

    let does_exist: bool = if row.0 >= 1 { true } else { false };

    if does_exist == false {
        sqlx::query("INSERT INTO users (username, password) VALUES ($1, $2)")
            .bind(&username)
            .bind(&password)
            .execute(&db_pool)
            .await
            .unwrap();
    } else {
        return (
            StatusCode::PAYMENT_REQUIRED,
            Json(json!({ "message" : "Username is already registered." })),
        );
    }

    (StatusCode::CREATED, Json(json!({ "username" : username })))
}

async fn authorize(db: impl PgExecutor<'_>, username: &str, password: &str) -> bool {
    if username.is_empty() || password.is_empty() {
        return false;
    }

    let row: Option<(String,)> = sqlx::query_as("SELECT password FROM users WHERE username = $1")
        .bind(&username)
        .fetch_optional(db)
        .await
        .unwrap();

    if let Some(val) = row {
        return password == val.0;
    }

    false
}

async fn auth_user(State(db_pool): State<PgPool>, headers: HeaderMap) -> (StatusCode, Json<Value>) {
    let username = headers["x-auth-user"].to_str().unwrap_or("");
    let password = headers["x-auth-key"].to_str().unwrap_or("");

    let mut tx = db_pool.begin().await.unwrap();

    if authorize(&mut *tx, username, password).await == true {
        return (StatusCode::OK, Json(json!({"authorized" : "OK"})));
    }

    tx.commit().await.unwrap();

    (
        StatusCode::UNAUTHORIZED,
        Json(json!({"message" : "Unauthorized"})),
    )
}

async fn update_progress(
    State(db_pool): State<PgPool>,
    headers: HeaderMap,
    Json(payload): Json<UpdateProgress>,
) -> StatusCode {
    let username = headers["x-auth-user"].to_str().unwrap_or("");
    let password = headers["x-auth-key"].to_str().unwrap_or("");

    let mut tx = db_pool.begin().await.unwrap();

    if authorize(&mut *tx, username, password).await == false {
        return StatusCode::UNAUTHORIZED;
    }

    tx.commit().await.unwrap();

    let timestamp = SystemTime::now()
        .duration_since(UNIX_EPOCH)
        .unwrap()
        .as_secs();

    let row: (i64,) = sqlx::query_as("SELECT id FROM users WHERE username = $1")
        .bind(&username)
        .fetch_one(&db_pool)
        .await
        .unwrap();

    let user_id = row.0;

    sqlx::query("INSERT INTO progresses (user_id, document, progress, percentage, device, device_id, timestamp) VALUES ($1, $2, $3, $4, $5, $6, $7) \
                    ON CONFLICT (user_id, document) DO UPDATE \
                    SET user_id = $1, document = $2, progress = $3, percentage = $4, device = $5, device_id = $6, timestamp = $7")
            .bind(user_id)
            .bind(&payload.document)
            .bind(&payload.progress)
            .bind(&payload.percentage)
            .bind(&payload.device)
            .bind(&payload.device_id)
            .bind(timestamp as i64)
            .execute(&db_pool)
            .await
            .unwrap();

    StatusCode::OK
}

async fn get_progress(
    State(db_pool): State<PgPool>,
    headers: HeaderMap,
    Path(document): Path<String>,
) -> (StatusCode, Json<Value>) {
    let username = headers["x-auth-user"].to_str().unwrap_or("");
    let password = headers["x-auth-key"].to_str().unwrap_or("");

    let mut tx = db_pool.begin().await.unwrap();

    if authorize(&mut *tx, username, password).await == false {
        return (StatusCode::UNAUTHORIZED, Json(json!("")));
    }

    tx.commit().await.unwrap();

    let row: (i64,) = sqlx::query_as("SELECT id FROM users WHERE username = $1")
        .bind(&username)
        .fetch_one(&db_pool)
        .await
        .unwrap();

    let user_id = row.0;

    let row: Option<(f32, String, String, String, i64)> = sqlx::query_as("SELECT percentage, progress, device, device_id, timestamp FROM progresses WHERE user_id = $1 AND document = $2")
        .bind(user_id)
        .bind(&document)
        .fetch_optional(&db_pool)
        .await
        .unwrap();

    if let Some(val) = row {
        let progress = GetProgress {
            percentage: val.0,
            progress: val.1,
            device: val.2,
            device_id: val.3,
            timestamp: val.4,
            document,
        };

        return (StatusCode::OK, Json(json!(progress)));
    }

    (StatusCode::OK, Json(json!("")))
}

async fn healthcheck() -> (StatusCode, Json<Value>) {
    (StatusCode::OK, Json(json!({ "state" : "OK" })))
}
